"""
This command-line tool (and Python package) allows for the easy extraction of
staves from musical scores. In particular, it makes it easy to extract the first
staff of any musical score that is in B&W format, and has a reasonable amount
of background noise.
"""

__version__ = "2021.3.1"
